// SPDX-License-Identifier: MIT
pragma solidity >=0.8.0 <0.9.0;

import "@openzeppelin/contracts/token/ERC20/utils/SafeERC20.sol";
import "../IGateway.sol";
import "../IDapp.sol";
import "../Utils.sol";
import "hardhat/console.sol";

contract HelloWorld is IDapp {
    using SafeERC20 for IERC20;
    IGateway public gatewayContract;
    string public greeting;
    uint256 public lastEventIdentifier;
    string public ackMessage;
    IERC20 public routeToken;
    address public assetVault;
    

    // for testing read query purposes
    uint256 public constant number = 20;

    event RequestFromRouterEvent(string indexed bridgeContract, bytes data);

    error CustomError(string message);

    constructor(address payable gatewayAddress, address routeTokenAddress, address assetVaultAddress) {
        gatewayContract = IGateway(gatewayAddress);
        routeToken = IERC20(routeTokenAddress);
        assetVault = assetVaultAddress;
    }

    function iSend(
        Utils.NativeTransferParams memory nativeTransferParams,
        string memory destChainId,
        bytes memory destContractAddress,
        uint256 gasLimit,
        uint256 relayerFees,
        Utils.AckType ackType,
        uint256 ackGasLimit
    ) external payable {
        routeToken.safeTransferFrom(msg.sender, address(this), nativeTransferParams.routeAmount);
        routeToken.safeIncreaseAllowance(assetVault, nativeTransferParams.routeAmount);
        
        bytes memory payload = abi.encode("Hello Router");

        bytes memory requestMetadata = abi.encode(
            gasLimit,
            0,
            ackGasLimit,
            0,
            relayerFees,
            ackType,
            false,
            ""
        );

        bytes memory requestPacket = abi.encode(payload, destContractAddress);

        uint256 gasBefore = gasleft();
        lastEventIdentifier = gatewayContract.iSend(1, nativeTransferParams, destChainId, requestMetadata, requestPacket);
        uint256 gasAfter = gasleft();

        console.log("Gas left: %s", gasBefore - gasAfter);
    }
    
    function iReceive(
        bytes memory srcContractAddress, 
        bytes memory payload, 
        string memory srcChainId
    ) external returns (bytes memory) {
        require(msg.sender == address(gatewayContract));

        string memory sampleStr = abi.decode(payload, (string));

        if (keccak256(bytes(sampleStr)) == keccak256(bytes(""))) {
            revert CustomError("String should not be empty");
        }

        greeting = sampleStr;
        return abi.encode(sampleStr);
    }

    function iAck(
        uint256 eventIdentifier,
        bool execFlag,
        bytes memory execData
    ) external {
        console.log(eventIdentifier);
        console.log(execFlag);
        console.logBytes(execData);

        bytes memory execDataa = abi.decode(execData, (bytes));
        ackMessage = abi.decode(execDataa, (string));    
    }

    receive() external payable {}

    function toAddress(bytes memory _bytes) internal pure returns (address contractAddress) {
        bytes20 srcTokenAddress;
        assembly {
            srcTokenAddress := mload(add(_bytes, 0x20))
        }
        contractAddress = address(srcTokenAddress);
    }

    function toBytes(address a) public pure returns (bytes memory b) {
        assembly {
            let m := mload(0x40)
            a := and(a, 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF)
            mstore(add(m, 20), xor(0x140000000000000000000000000000000000000000, a))
            mstore(0x40, add(m, 52))
            b := m
        }
    }
}
